(ns ^{:doc "Definition of the StatusMessage protocol"}
 status-im.transport.message.core
  (:require [status-im.chat.models.message :as models.message]
            [status-im.chat.models :as models.chat]
            [status-im.chat.models.reactions :as models.reactions]
            [status-im.contact.core :as models.contact]
            [status-im.communities.core :as models.communities]
            [status-im.pairing.core :as models.pairing]
            [status-im.transport.filters.core :as models.filters]
            [status-im.data-store.reactions :as data-store.reactions]
            [status-im.data-store.contacts :as data-store.contacts]
            [status-im.data-store.chats :as data-store.chats]
            [status-im.data-store.invitations :as data-store.invitations]
            [status-im.data-store.activities :as data-store.activities]
            [status-im.group-chats.core :as models.group]
            [status-im.utils.fx :as fx]
            [status-im.utils.types :as types]
            [status-im.constants :as constants]
            [status-im.multiaccounts.model :as multiaccounts.model]
            [status-im.notifications-center.core :as notifications-center]
            [clojure.string :as string]))

(fx/defn process-next
  [cofx ^js response-js sync-handler]
  (if sync-handler
    (sync-handler cofx response-js true)
    {:utils/dispatch-later [{:ms 20 :dispatch [:process-response response-js]}]}))

(fx/defn process-response
  {:events [:process-response]}
  [{:keys [db] :as cofx} ^js response-js process-async]
  (let [^js communities (.-communities response-js)
        ^js requests-to-join-community (.-requestsToJoinCommunity response-js)
        ^js chats (.-chats response-js)
        ^js contacts (.-contacts response-js)
        ^js installations (.-installations response-js)
        ^js messages (.-messages response-js)
        ^js emoji-reactions (.-emojiReactions response-js)
        ^js filters (.-filters response-js)
        ^js removed-filters (.-removedFilters response-js)
        ^js invitations (.-invitations response-js)
        ^js removed-chats (.-removedChats response-js)
        ^js activity-notifications (.-activityCenterNotifications response-js)
        sync-handler (when-not process-async process-response)]

    (cond

      (seq chats)
      (do
        (js-delete response-js "chats")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.chat/ensure-chats (map #(-> %
                                                      (data-store.chats/<-rpc)
                                                      ;;TODO why here?
                                                      (dissoc :unviewed-messages-count))
                                                 (types/js->clj chats)))))

      (seq messages)
      (models.message/receive-many cofx response-js)

      (seq activity-notifications)
      (do
        (js-delete response-js "activityCenterNotifications")
        (fx/merge cofx
                  (notifications-center/handle-activities (map data-store.activities/<-rpc
                                                               (types/js->clj activity-notifications)))
                  (process-next response-js sync-handler)))

      (seq installations)
      (let [installations-clj (types/js->clj installations)]
        (js-delete response-js "installations")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.pairing/handle-installations installations-clj)))

      (seq contacts)
      (let [contacts-clj (types/js->clj contacts)]
        (js-delete response-js "contacts")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.contact/ensure-contacts (map data-store.contacts/<-rpc contacts-clj))))

      (seq communities)
      (let [communities-clj (types/js->clj communities)]
        (js-delete response-js "communities")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.communities/handle-communities (types/js->clj communities-clj))))

      (seq removed-chats)
      (let [removed-chats-clj (types/js->clj removed-chats)]
        (js-delete response-js "removedChats")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.communities/handle-removed-chats (types/js->clj removed-chats-clj))))

      (seq requests-to-join-community)
      (let [request (.pop requests-to-join-community)]
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.communities/handle-request-to-join (types/js->clj request))))

      (seq emoji-reactions)
      (let [reactions (types/js->clj emoji-reactions)]
        (js-delete response-js "emojiReactions")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.reactions/receive-signal (map data-store.reactions/<-rpc reactions))))

      (seq invitations)
      (let [invitations (types/js->clj invitations)]
        (js-delete response-js "invitations")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.group/handle-invitations (map data-store.invitations/<-rpc invitations))))
      (seq filters)
      (let [filters (types/js->clj filters)]
        (js-delete response-js "filters")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.filters/handle-filters filters)))

      (seq removed-filters)
      (let [removed-filters (types/js->clj removed-filters)]
        (js-delete response-js "removedFilters")
        (fx/merge cofx
                  (process-next response-js sync-handler)
                  (models.filters/handle-filters-removed filters))))))

(defn group-by-and-update-unviewed-counts
  "group messages by current chat, profile updates, transactions and update unviewed counters in db for not curent chats"
  [{:keys [current-chat-id db] :as acc} ^js message-js]
  (let [chat-id (.-localChatId message-js)
        message-type (.-messageType message-js)
        from (.-from message-js)
        new (.-new message-js)
        current (= current-chat-id chat-id)
        profile (models.chat/profile-chat? {:db db} chat-id)
        tx-hash (and (.-commandParameters message-js) (.-commandParameters.transactionHash message-js))]
    (cond-> acc
      current
      (update :messages conj message-js)

      profile
      (update :statuses conj message-js)

      ;;update counter
      (and (not current)
           new
           (not profile)
           (not (= message-type constants/message-type-private-group-system-message))
           (not (= from (multiaccounts.model/current-public-key {:db db}))))
      (update-in [:db :chats chat-id :unviewed-messages-count] inc)

      ;;conj incoming transaction for :watch-tx
      (not (string/blank? tx-hash))
      (update :transactions conj tx-hash)

      :always
      (update :chats conj chat-id))))

(defn sort-js-messages!
  "sort messages, so we can start process latest first,in that case we only need to process frist 20 and drop others"
  [response-js messages]
  (if (seq messages)
    (set! (.-messages response-js)
          (.sort (to-array messages)
                 (fn [a b]
                   (- (.-clock b) (.-clock a)))))
    (js-delete response-js "messages")))

(fx/defn sanitize-messages-and-process-response
  "before processing we want to filter and sort messages, so we can process first only messages which will be showed"
  {:events [:sanitize-messages-and-process-response]}
  [{:keys [db] :as cofx} ^js response-js process-async]
  (let [current-chat-id (:current-chat-id db)
        {:keys [db messages transactions chats statuses]}
        (reduce group-by-and-update-unviewed-counts
                {:db db :chats #{} :transactions #{} :statuses [] :messages []
                 :current-chat-id current-chat-id}
                (.-messages response-js))]
    (sort-js-messages! response-js messages)
    (fx/merge cofx
              {:db db
               :utils/dispatch-later (concat []
                                             (when (seq statuses)
                                               [{:ms 100 :dispatch [:process-statuses statuses]}])
                                             (when (seq transactions)
                                               (for [transaction-hash transactions]
                                                 {:ms 100 :dispatch [:watch-tx transaction-hash]}))
                                             (when (seq chats)
                                               [{:ms 100 :dispatch [:chat/join-times-messages-checked chats]}]))}
              (process-response response-js process-async))))

(fx/defn remove-hash
  [{:keys [db]} envelope-hash]
  {:db (update db :transport/message-envelopes dissoc envelope-hash)})

(fx/defn check-confirmations
  [{:keys [db] :as cofx} status chat-id message-id]
  (when-let [{:keys [pending-confirmations not-sent]}
             (get-in db [:transport/message-ids->confirmations message-id])]
    (if (zero? (dec pending-confirmations))
      (fx/merge cofx
                {:db (update db
                             :transport/message-ids->confirmations
                             dissoc message-id)}
                (models.message/update-message-status chat-id
                                                      message-id
                                                      (if not-sent
                                                        :not-sent
                                                        status))
                (remove-hash message-id))
      (let [confirmations {:pending-confirmations (dec pending-confirmations)
                           :not-sent  (or not-sent
                                          (= :not-sent status))}]
        {:db (assoc-in db
                       [:transport/message-ids->confirmations message-id]
                       confirmations)}))))

(fx/defn update-envelope-status
  [{:keys [db] :as cofx} message-id status]
  (if-let [{:keys [chat-id]}
           (get-in db [:transport/message-envelopes message-id])]
    (when-let [{:keys [from]} (get-in db [:messages chat-id message-id])]
      (check-confirmations cofx status chat-id message-id))
    ;; We don't have a message-envelope for this, might be that the confirmation
    ;; came too early
    {:db (update-in db [:transport/message-confirmations message-id] conj status)}))

(fx/defn update-envelopes-status
  [{:keys [db] :as cofx} message-id status]
  (apply fx/merge cofx (map #(update-envelope-status % status) message-id)))

(fx/defn set-message-envelope-hash
  "message-type is used for tracking"
  [{:keys [db] :as cofx} chat-id message-id message-type]
  ;; Check first if the confirmation has already arrived
  (let [statuses (get-in db [:transport/message-confirmations message-id])
        check-confirmations-fx (map
                                #(check-confirmations % chat-id message-id)
                                statuses)

        add-envelope-data (fn [{:keys [db]}]
                            {:db (-> db
                                     (update :transport/message-confirmations dissoc message-id)
                                     (assoc-in [:transport/message-envelopes message-id]
                                               {:chat-id      chat-id
                                                :message-type message-type})
                                     (update-in [:transport/message-ids->confirmations message-id]
                                                #(or % {:pending-confirmations 1})))})]
    (apply fx/merge cofx (conj check-confirmations-fx add-envelope-data))))

(fx/defn transport-message-sent
  {:events [:transport/message-sent]}
  [cofx response-js]
  (let [set-hash-fxs (map (fn [{:keys [localChatId id messageType]}]
                            (set-message-envelope-hash localChatId id messageType))
                          (types/js->clj (.-messages response-js)))]
    (apply fx/merge cofx
           (conj set-hash-fxs
                 #(sanitize-messages-and-process-response % response-js false)))))
